@@ -17,12 +17,14 @@ try:
 except ImportError:
     from backports.shutil_which import which
 
-COMMANDS = ["show", "diff", "merge", "diff-web", "merge-web", "mergetool"]
+COMMANDS = ["show", "diff", "merge", "diff-web", "merge-web", "mergetool",
+            "config-git", "reg-extensions"]
 HELP_MESSAGE_VERBOSE = ("Usage: nbdime [OPTIONS]\n\n"
                        "OPTIONS: -h, --version, COMMANDS{%s}\n\n"
                        "Examples: nbdime --version\n"
                        "          nbdime show -h\n"
                        "          nbdime merge-web\n\n"
+                       "          nbdime config-git (--enable | --disable)"
                        "Documentation: https://nbdime.readthedocs.io" % ", ".join(COMMANDS))
 
 
@@ -65,6 +67,28 @@ def main_dispatch(args=None):
         from nbdime.webapp.nbmergeweb import main
     elif cmd == 'mergetool':
         main = main_mergetool
+    elif cmd == 'server':
+        from nbdime.webapp.nbdimeserver import main
+    elif cmd == 'config-git':
+        # Call all git configs
+        from nbdime.vcs.git.diffdriver import main as diff_driver
+        from nbdime.vcs.git.difftool import main as diff_tool
+        from nbdime.vcs.git.mergedriver import main as merge_driver
+        from nbdime.vcs.git.mergetool import main as merge_tool
+        args = ['config'] + args
+        # Short-circuit on first non-zero return code:
+        return (
+            diff_driver(args) or
+            merge_driver(args) or
+            diff_tool(args) or
+            merge_tool(args)
+        )
+    elif cmd == 'reg-extensions':
+        # Register nbdime extensions
+        call('jupyter serverextension enable --py nbdime'.split() + args)
+        call('jupyter nbextension install --py nbdime'.split() + args)
+        call('jupyter nbextension enable --py nbdime'.split() + args)
+        return 0
     else:
         if cmd == '--version':
             sys.exit(__version__)
