@@ -6,20 +6,16 @@
 from __future__ import unicode_literals
 from __future__ import print_function
 
-import io
 import os
 import sys
 import argparse
 import nbformat
-import json
 
-import nbdime
-from nbdime.prettyprint import pretty_print_notebook
-from nbdime.args import add_generic_args, add_filename_args
+from nbdime.prettyprint import pretty_print_notebook, PrettyPrintConfig
+from nbdime.args import add_generic_args, IgnorableAction, process_exclusive_ignorables
 from nbdime.utils import setup_std_streams
 
 
-
 _description = """Show a Jupyter notebook in terminal.
 By default shows all notebook fields.
 Limit to specific fields by passing options.
@@ -49,17 +45,16 @@ def main_show(args):
         class Printer:
             def write(self, text):
                 print(text, end="")
-        if not any((args.sources, args.outputs, args.attachments, args.metadata, args.details)):
-            ppargs = None
-        else:
-            ppargs = args
+
+        # This configures which parts to include/ignore
+        config = PrettyPrintConfig(out=Printer(), include=args)
 
         if len(args.notebook) > 1:
             # 'more' prints filenames with colons, should be good enough for us as well
             print(":"*14)
             print(fn)
             print(":"*14)
-        pretty_print_notebook(nb, ppargs, Printer())
+        pretty_print_notebook(nb, config)
 
     return 0
 
@@ -76,28 +71,23 @@ def _build_arg_parser():
     # Things we can choose to show or not
     parser.add_argument(
         '-s', '--sources',
-        action="store_true",
-        default=False,
+        action=IgnorableAction,
         help="show sources.")
     parser.add_argument(
         '-o', '--outputs',
-        action="store_true",
-        default=False,
+        action=IgnorableAction,
         help="show outputs.")
     parser.add_argument(
         '-a', '--attachments',
-        action="store_true",
-        default=False,
+        action=IgnorableAction,
         help="show attachments.")
     parser.add_argument(
         '-m', '--metadata',
-        action="store_true",
-        default=False,
+        action=IgnorableAction,
         help="show metadata.")
     parser.add_argument(
         '-d', '--details',
-        action="store_true",
-        default=False,
+        action=IgnorableAction,
         help="show details not covered by other options.")
 
     return parser
@@ -108,7 +98,9 @@ def main(args=None):
         args = sys.argv[1:]
     setup_std_streams()
     arguments = _build_arg_parser().parse_args(args)
-    nbdime.log.init_logging(level=arguments.log_level)
+    process_exclusive_ignorables(
+        arguments,
+        ('sources', 'outputs', 'attachments', 'metadata', 'details'))
     return main_show(arguments)
 
 
