@@ -7,7 +7,7 @@ import io
 import json
 import logging
 import os
-from subprocess import CalledProcessError, Popen, check_call
+from subprocess import CalledProcessError, check_call
 import sys
 import time
 
@@ -17,7 +17,10 @@ from tornado.httputil import url_concat
 
 import nbformat
 
-from .utils import assert_clean_exit, get_output, call
+from .utils import (
+    assert_clean_exit, get_output, call, WEB_TEST_TIMEOUT,
+    write_local_hg_config, wait_up
+)
 
 import nbdime
 from nbdime.nbshowapp import main_show
@@ -29,13 +32,19 @@ from nbdime import (
     nbdiffapp,
     nbmergeapp,
     nbpatchapp,
-    gitdiffdriver,
-    gitdifftool,
-    gitmergedriver,
-    gitmergetool,
 )
-import nbdime.webapp.nbdiffweb
-import nbdime.webapp.nbmergeweb
+from nbdime.vcs.git import (
+    diffdriver as gitdiffdriver,
+    difftool as gitdifftool,
+    mergedriver as gitmergedriver,
+    mergetool as gitmergetool,
+)
+from nbdime.vcs.hg import (
+    diff as hgdiff,
+    diffweb as hgdiffweb,
+    merge as hgmerge,
+    mergeweb as hgmergeweb,
+)
 from nbdime.utils import EXPLICIT_MISSING_FILE
 
 
@@ -79,6 +88,26 @@ def test_nbdiff_app_null_file(filespath):
     assert 0 == main_diff(args)
 
 
+def test_nbdiff_app_gitrefs(git_repo2):
+    args = nbdiffapp._build_arg_parser().parse_args(['local', 'remote'])
+    assert 0 == main_diff(args)
+
+    args = nbdiffapp._build_arg_parser().parse_args(['local', 'remote', 'sub/subfile.ipynb'])
+    assert 0 == main_diff(args)
+
+    args = nbdiffapp._build_arg_parser().parse_args(['local', 'remote', 'sub/subfile.ipynb', 'diff.ipynb'])
+    assert 0 == main_diff(args)
+
+    args = nbdiffapp._build_arg_parser().parse_args(['local', 'sub/subfile.ipynb', 'diff.ipynb'])
+    assert 0 == main_diff(args)
+
+    args = nbdiffapp._build_arg_parser().parse_args(['sub/subfile.ipynb'])
+    assert 0 == main_diff(args)
+
+    args = nbdiffapp._build_arg_parser().parse_args([])
+    assert 0 == main_diff(args)
+
+
 def test_nbdiff_app_unicode_safe(filespath):
     afn = os.path.join(filespath, "unicode--1.ipynb")
     bfn = os.path.join(filespath, "unicode--2.ipynb")
@@ -88,6 +117,47 @@ def test_nbdiff_app_unicode_safe(filespath):
     check_call([sys.executable, '-m', 'nbdime.nbdiffapp', afn, bfn], env=env)
 
 
+def test_nbdiff_app_only_source(filespath, tmpdir, reset_diff_targets):
+    afn = os.path.join(filespath, "multilevel-test-base.ipynb")
+    bfn = os.path.join(filespath, "multilevel-test-local.ipynb")
+    dfn = os.path.join(tmpdir.dirname, "diff_output.json")
+
+    args = nbdiffapp._build_arg_parser().parse_args([afn, bfn, '--out', dfn, '-s'])
+    assert 0 == main_diff(args)
+    with io.open(dfn) as df:
+        diff = json.load(df)
+    for key in ('cells', 1, 'source'):
+        assert len(diff) == 1
+        assert diff[0]['key'] == key
+        assert diff[0]['op'] == 'patch'
+        diff = diff[0]['diff']
+
+
+def test_nbdiff_app_ignore_source(filespath, tmpdir, reset_diff_targets):
+    afn = os.path.join(filespath, "multilevel-test-base.ipynb")
+    bfn = os.path.join(filespath, "multilevel-test-local.ipynb")
+    dfn = os.path.join(tmpdir.dirname, "diff_output.json")
+
+    args = nbdiffapp._build_arg_parser().parse_args([afn, bfn, '--out', dfn, '-S'])
+    assert 0 == main_diff(args)
+    with io.open(dfn) as df:
+        diff = json.load(df)
+    for key in ('cells', 2, 'outputs'):
+        assert len(diff) == 1
+        assert diff[0]['key'] == key
+        assert diff[0]['op'] == 'patch'
+        diff = diff[0]['diff']
+
+
+def test_nbdiff_app_color_words(filespath):
+    # Simply check that the --color-words argument is accepted, not behavior
+    afn = os.path.join(filespath, "multilevel-test-base.ipynb")
+    bfn = os.path.join(filespath, "multilevel-test-local.ipynb")
+
+    args = nbdiffapp._build_arg_parser().parse_args([afn, bfn, '--color-words'])
+    assert 0 == main_diff(args)
+
+
 def test_nbmerge_app(tempfiles, capsys):
     bfn = os.path.join(tempfiles, "multilevel-test-base.ipynb")
     lfn = os.path.join(tempfiles, "multilevel-test-local.ipynb")
@@ -102,7 +172,7 @@ def test_nbmerge_app(tempfiles, capsys):
 
     nb_stdout, err = capsys.readouterr()
 
-    assert 0 == nbmergeapp.main([bfn, lfn, rfn, '-o', ofn])
+    assert 0 == nbmergeapp.main([bfn, lfn, rfn, '--out', ofn])
     out, err = capsys.readouterr()
     # no stdout when sending output to file
     assert out == ''
@@ -159,7 +229,7 @@ def test_nbmerge_app_conflict(tempfiles, capsys):
     assert 1 == nbmergeapp.main([bfn, lfn, rfn])
     nb_stdout, err = capsys.readouterr()
 
-    assert 1 == nbmergeapp.main([bfn, lfn, rfn, '-o', ofn])
+    assert 1 == nbmergeapp.main([bfn, lfn, rfn, '--out', ofn])
     out, err = capsys.readouterr()
     # no stdout when sending output to file
     assert out == ''
@@ -178,7 +248,7 @@ def test_nbmerge_app_decisions(tempfiles, capsys, reset_log):
     rfn = os.path.join(tempfiles, "inline-conflict--3.ipynb")
     ofn = os.path.join(tempfiles, "inline-conflict-out.ipynb")
 
-    assert 1 == nbmergeapp.main([bfn, lfn, rfn, '--decisions', '-o', ofn])
+    assert 1 == nbmergeapp.main([bfn, lfn, rfn, '--decisions', '--out', ofn])
     out, err = capsys.readouterr()
     # decisions are logged to stderr:
     assert 'conflicted decisions' in err
@@ -189,7 +259,7 @@ def test_nbmerge_app_decisions(tempfiles, capsys, reset_log):
 
 
 def test_diffdriver_config(git_repo):
-    main = nbdime.gitdiffdriver.main
+    main = gitdiffdriver.main
     with assert_clean_exit():
         main(['config', '-h'])
     assert not os.path.exists('.gitattributes')
@@ -216,7 +286,7 @@ def _check_diffdriver_disabled():
 
 
 def test_difftool_config(git_repo):
-    main = nbdime.gitdifftool.main
+    main = gitdifftool.main
 
     with assert_clean_exit():
         main(['config', '-h'])
@@ -247,7 +317,7 @@ def _check_difftool_disabled():
 
 
 def test_mergedriver_config(git_repo):
-    main = nbdime.gitmergedriver.main
+    main = gitmergedriver.main
     with assert_clean_exit():
         main(['config', '-h'])
     assert not os.path.exists('.gitattributes')
@@ -274,7 +344,7 @@ def _check_mergedriver_disabled():
 
 
 def test_mergetool_config(git_repo):
-    main = nbdime.gitmergetool.main
+    main = gitmergetool.main
     with assert_clean_exit():
         main(['config', '-h'])
 
@@ -335,19 +405,19 @@ def test_config_git_fails(git_repo):
         _check_mergetool_disabled()
 
 
-def test_diffdriver(git_repo):
-    nbdime.gitdiffdriver.main(['config', '--enable'])
+def test_git_diffdriver(git_repo):
+    gitdiffdriver.main(['config', '--enable'])
     out = get_output('git diff base diff.ipynb')
     assert 'nbdiff' in out
 
 
-def test_mergedriver(git_repo, filespath):
+def test_git_mergedriver(git_repo, filespath):
     # enable diff/merge drivers
-    nbdime.gitdiffdriver.main(['config', '--enable'])
-    nbdime.gitmergedriver.main(['config', '--enable'])
+    gitdiffdriver.main(['config', '--enable'])
+    gitmergedriver.main(['config', '--enable'])
     # run merge with no conflicts
     out = get_output('git merge remote-no-conflict', err=True)
-    assert 'nbmergeapp' in out
+    assert 'Auto-merging merge-no-conflict.ipynb' in out
     with open('merge-no-conflict.ipynb') as f:
         merged = f.read()
 
@@ -373,29 +443,13 @@ def test_mergedriver(git_repo, filespath):
     nbformat.validate(nb)
 
 
-WEB_TEST_TIMEOUT = 15
-
-
-def _wait_up(url, interval=0.1, check=None):
-    while True:
-        try:
-            r = requests.get(url)
-        except Exception as e:
-            if check:
-                assert check()
-            print("waiting for %s" % url)
-            time.sleep(interval)
-        else:
-            break
-
-
-@pytest.mark.timeout(timeout=WEB_TEST_TIMEOUT)
-def test_difftool(git_repo, request):
-    nbdime.gitdifftool.main(['config', '--enable'])
+@pytest.mark.timeout(timeout=3*WEB_TEST_TIMEOUT)
+def test_git_difftool(git_repo, unique_port, popen_with_terminator):
+    gitdifftool.main(['config', '--enable'])
     cmd = get_output('git config --get --local difftool.nbdime.cmd').strip()
 
     # pick a non-random port so we can connect later, and avoid opening a browser
-    port = 62021
+    port = unique_port
     cmd = cmd + ' --port=%i --browser=disabled' % port
     call(['git', 'config', 'difftool.nbdime.cmd', cmd])
 
@@ -403,19 +457,12 @@ def test_difftool(git_repo, request):
     with open('.gitattributes', 'w') as f:
         f.write('*.ipynb\tdiff=notnbdime')
 
-    process = Popen(['git', 'difftool', '--tool=nbdime', 'base'])
-
-    def _term():
-        try:
-            process.terminate()
-        except OSError:
-            pass
-    request.addfinalizer(_term)
+    process = popen_with_terminator(['git', 'difftool', '--tool=nbdime', 'base'])
 
     # 3 is the number of notebooks in this diff
     url = 'http://127.0.0.1:%i' % port
     for i in range(3):
-        _wait_up(url, check=lambda: process.poll() is None)
+        wait_up(url, check=lambda: process.poll() is None)
         # server started
         r = requests.get(url + '/difftool')
         r.raise_for_status()
@@ -428,31 +475,121 @@ def test_difftool(git_repo, request):
     assert process.poll() == 0
 
 
-@pytest.mark.timeout(timeout=WEB_TEST_TIMEOUT)
-def test_mergetool(git_repo, request):
-    nbdime.gitmergetool.main(['config', '--enable'])
+@pytest.mark.timeout(timeout=3*WEB_TEST_TIMEOUT)
+def test_git_mergetool(git_repo, unique_port, popen_with_terminator):
+    gitmergetool.main(['config', '--enable'])
     cmd = get_output('git config --get --local mergetool.nbdime.cmd').strip()
 
     # pick a non-random port so we can connect later, and avoid opening a browser
-    port = 62022
+    port = unique_port
     cmd = cmd + ' --port=%i --browser=disabled' % port
     call(['git', 'config', 'mergetool.nbdime.cmd', cmd])
     call(['git', 'config', 'mergetool.nbdime.trustExitCode', 'true'])
 
     with pytest.raises(CalledProcessError):
         call('git merge remote-conflict')
-    process = Popen(['git', 'mergetool', '--no-prompt', '--tool=nbdime', 'merge-conflict.ipynb'])
+    process = popen_with_terminator([
+        'git', 'mergetool', '--no-prompt', '--tool=nbdime', 'merge-conflict.ipynb'])
 
-    def _term():
-        try:
-            process.terminate()
-        except OSError:
-            pass
-    request.addfinalizer(_term)
-
-    # 3 is the number of notebooks in this diff
+    # 3 total web calls: mergetool, api/store, api/closetool
     url = 'http://127.0.0.1:%i' % port
-    _wait_up(url, check=lambda: process.poll() is None)
+    wait_up(url, check=lambda: process.poll() is None)
+    # server started
+    r = requests.get(url + '/mergetool')
+    r.raise_for_status()
+    r = requests.post(
+        url_concat(url + '/api/store', {'outputfilename': 'merge-conflict.ipynb'}),
+        data=json.dumps({
+            'merged': nbformat.v4.new_notebook(),
+        })
+    )
+    r.raise_for_status()
+    # close it
+    r = requests.post(url + '/api/closetool', headers={'exit_code': '0'})
+    r.raise_for_status()
+    # wait for exit
+    process.wait()
+    assert process.poll() == 0
+
+
+def test_hg_diff(hg_repo):
+    write_local_hg_config(hg_repo)
+    # ExtDiff for some reason always returns 1.
+    out = get_output('hg nbdiff -r base:local diff.ipynb', returncode=1)
+    assert 'nbdiff' in out
+
+
+def test_hg_mergedriver(hg_repo, filespath):
+    # enable diff/merge drivers
+    write_local_hg_config(hg_repo)
+    # run merge with no conflicts
+    out = get_output('hg merge remote-no-conflict', err=True)
+    assert 'nbmergeapp' in out
+    with open('merge-no-conflict.ipynb') as f:
+        merged = f.read()
+
+    with open(os.path.join(filespath, 'multilevel-test-merged.ipynb')) as f:
+        expected = f.read()
+
+    # verify merge success
+    assert merged == expected
+
+    # reset
+    call('hg update --clean local')
+
+    # run merge with conflicts
+    with pytest.raises(CalledProcessError):
+        call('hg merge remote-conflict')
+
+    status = get_output('hg status')
+    assert 'merge-conflict.ipynb' in status
+    out = get_output('hg nbdiff', returncode=1)
+    assert 'nbdiff' in out
+    # verify that the conflicted result is a valid notebook
+    nb = nbformat.read('merge-conflict.ipynb', as_version=4)
+    nbformat.validate(nb)
+
+
+
+@pytest.mark.timeout(timeout=3*WEB_TEST_TIMEOUT)
+def test_hg_diffweb(hg_repo, unique_port, popen_with_terminator):
+    # enable diff/merge drivers
+    write_local_hg_config(hg_repo)
+
+    process = popen_with_terminator([
+        'hg', 'nbdiffweb', '-r', 'base', '-o', '--port=%i' % unique_port])
+
+    # 3 is the number of notebooks in this diff
+    url = 'http://127.0.0.1:%i' % unique_port
+    wait_up(url, check=lambda: process.poll() is None)
+    for i in range(3):
+        # server started
+        r = requests.get(url + '/difftool')
+        r.raise_for_status()
+        # close it
+        r = requests.post(url + '/api/closetool', headers={'exit_code': '0'})
+        r.raise_for_status()
+        time.sleep(0.25)
+    # wait for exit
+    process.wait()
+    assert process.poll() == 1  # hg ExtDiff returns 1 for some reason
+
+
+@pytest.mark.timeout(timeout=WEB_TEST_TIMEOUT)
+def test_hg_mergetool(hg_repo, unique_port, popen_with_terminator):
+    # enable diff/merge drivers
+    write_local_hg_config(hg_repo)
+
+    with pytest.raises(CalledProcessError):
+        call('hg merge remote-conflict')
+    config_override = '--log-level DEBUG --browser=disabled --port=%d $base $local $other $output' % unique_port
+    process = popen_with_terminator([
+        'hg', 'resolve', '--tool', 'nbdimeweb',
+        '--config', 'merge-tools.nbdimeweb.args=%s' % config_override,
+        'merge-conflict.ipynb'])
+
+    url = 'http://127.0.0.1:%i' % unique_port
+    wait_up(url, check=lambda: process.poll() is None)
     # server started
     r = requests.get(url + '/mergetool')
     r.raise_for_status()
